/*
 * hostname caching procedures for WWW proxies
 *      by k-chinen@is.aist-nara.ac.jp, 1999, 2000
 *
 * $Id$
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <limits.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <assert.h>
#include <errno.h>


#include "range.h"
#include "socket.h"
#include "misc.h"
#include "sim_stat.h"
#include "msg.h"
#include "trace.h"

#ifdef DEFAULT_HRES_SERV_PORT
#define HOSTQUERY_PORT  DEFAULT_HRES_SERV_PORT
#else
#ifndef HOSTQUERY_PORT
#define HOSTQUERY_PORT  (1245)
#endif
#endif

#define STAT_LIFETIME   (1024*1024*1024)
#define STAT_REPORT     (20)
#define STAT_LEN        (1024)


#ifndef HCACHE_NSLOT
#define HCACHE_NSLOT       (79)
#endif


#define _TRACE_HCACHE
#define _TRACE_DETAIL_HCACHE


TRACE_REGIST_NIC("hcache");

/* static */ sim_stat_rec hres_stat;
long hres_stat_count[STAT_LEN];

struct sockaddr_in resolver_addr;


pthread_mutex_t hcache_lock = PTHREAD_MUTEX_INITIALIZER;
char hname_slot[HCACHE_NSLOT][L_HOST];
struct sockaddr_in haddr_slot[HCACHE_NSLOT];
int slot_use[HCACHE_NSLOT];


static
unsigned
int
hhash(unsigned char *b)
{
    register unsigned int sum;
    register unsigned char *p, *q;

    sum = 0;
    p = b;
    q = "dUmMy";
    while(*p) {
        sum = (sum<<1) + (*p) * (*q);
    q = p;
        p++;
    }
    if(sum<0) {
    sum = -sum;
    }

    return sum % HCACHE_NSLOT;
}

void
hcache_init()
{
    TRACE_RECEIPT;
}

int
hcache_setup()
{
    int chk;
    int i;

    Trace("parameters for hostname caching\n");
    Trace("    HOSTQUERY_PORT   = %d\n", HOSTQUERY_PORT);
    Trace("    HCACHE_NSLOT     = %d\n", HCACHE_NSLOT);
    Trace("\n");


    chk = addr_byname("localhost", HOSTQUERY_PORT, &resolver_addr);
    if(chk) {
        Log("; resolver not found");
        return -1;
    }

    sim_stat_alloc_new(&hres_stat, "hostname resolve time",
        STAT_LEN, hres_stat_count);

    MUTEX_LOCK(&hcache_lock);
    for(i=0;i<HCACHE_NSLOT;i++) {
        hname_slot[i][0] = '\0';
        slot_use[i] = 0;
    }
    MUTEX_UNLOCK(&hcache_lock);

    return 0;
}

void
hcache_list()
{
    int i;

    MUTEX_LOCK(&hcache_lock);
    fflush(stdout);
    fprintf(stderr, "=== hostname cache ===\n");
    for(i=0;i<HCACHE_NSLOT;i++) {
        fprintf(stderr, "%4d: %12d '%s'\n", i, slot_use[i], hname_slot[i]);
    }
    fflush(stderr);
    MUTEX_UNLOCK(&hcache_lock);
}



int
hcache_got_addr_direct(char *name, struct sockaddr_in* addr)
{
    int nr, nw;
    int chk;
    int ret;
    char protocol[L_URL], portstr[L_URL];
    int qfd;
    char addrbuf[32];

#ifdef TRACE_HCACHE
    Trace("hcache_got_addr_direct: name '%s'\n", name);
#endif

    if(name==NULL || *name=='\0') {
        return -1;
    }
#ifdef TRACE_HCACHE
    Trace("hcache_got_addr_direct: name '%s'\n", name);
#endif


    ret = 0;

    if(*name>='0' && *name<='9') {
        register char *p;

        /* check, name is filled digits and/or dots */
        p = name;
        while(*p && ((*p>='0' && *p<='9') || *p=='.')) {
            p++;
        }

        if(!*p) {
            memset(addr, 0, sizeof(addr));
            addr->sin_family = AF_INET;
            addr->sin_port   = 0;
            addr->sin_addr.s_addr = inet_addr(name);

            goto out;
        }
    }


    /***
     *** name resolving
     ***/
    qfd = TCP_connect_addr(&resolver_addr, HOSTQUERY_PORT);
    if(qfd<0) {
        Error("fail TCP_connect_addr() ret %d (%d)", qfd, errno);
        ret = -2;
        goto out;
    }
#ifdef TRACE_HCACHE
    Trace("qfd %d, query host-'%s' %d\n",
        qfd, name, strlen(name)+1);
#endif

try_write:
    nw = write(qfd, name, strlen(name)+1);
#ifdef TRACE_HCACHE
    Trace("write %d bytes (%d)\n", nw, errno);
#endif
    if(nw<0) {
        if(errno==EINTR) {
            goto try_write;
        }
        Error("write() ret %d (%d)", nw, errno);
        ret = -3;
        goto try_close;
    }

try_read:
    nr = read(qfd, addrbuf, 32);
#ifdef TRACE_HCACHE
    Trace("read %d bytes (%d)\n", nr, errno);
#endif
    if(nr<0) {
        if(errno==EINTR) {
            goto try_read;
        }
        Error("read() ret %d (%d)", nr, errno);
        ret = -4;
        goto try_close;
    }
    else
    if(nr>0) {
        addr->sin_family = AF_INET;
        memcpy(&(addr->sin_addr.s_addr), addrbuf, 4);
    }
    else {
        ret = -5;
    }

try_close:
    shutdown(qfd, 2);
    close(qfd);

out:
    return ret;
}


int
hcache_send_query_addr_direct(char *name, struct sockaddr_in* addr, int *qfd)
{
    int nr, nw;
    int chk;
    int ret;
    char protocol[L_URL], portstr[L_URL];
    char addrbuf[32];

#ifdef TRACE_HCACHE
    Trace("hcache_send_query_addr_direct: name '%s'\n", name);
#endif

    if(name==NULL || *name=='\0') {
        return -1;
    }

    ret = 0;

    if(*name>='0' && *name<='9') {
        register char *p;

        /* check, name is filled digits and/or dots */
        p = name;
        while(*p && ((*p>='0' && *p<='9') || *p=='.')) {
            p++;
        }

        if(!*p) {
            memset(addr, 0, sizeof(addr));
            addr->sin_family = AF_INET;
            addr->sin_port   = 0;
            addr->sin_addr.s_addr = inet_addr(name);

            goto out;
        }
    }

    /***
     *** name resolving
     ***/
    *qfd = TCP_connect_addr(&resolver_addr, HOSTQUERY_PORT);
    if(*qfd<0) {
        Error("fail TCP_connect_addr() ret %d (%d)", *qfd, errno);
        ret = -2;
        goto out;
    }
#ifdef TRACE_HCACHE
    Trace("  query: fd %d, host '%s', length %d\n",
        *qfd, name, strlen(name)+1);
#endif

try_write:
    nw = write(*qfd, name, strlen(name)+1);
#ifdef TRACE_HCACHE
    Trace("  write %d bytes (%d)\n", nw, errno);
#endif
    if(nw<0) {
        if(errno==EINTR) {
            goto try_write;
        }
        Error("write() ret %d (%d)", nw, errno);
        ret = -3;

        shutdown(*qfd, 2);
        close(*qfd);
        *qfd = -1;

        goto out;
    }

out:
    return ret;
}

/*
 * success  0
 * fail     !0
 *
 */
int
hcache_recv_response_addr_direct(char *name, struct sockaddr_in* addr, int *qfd)
{
    int nr, nw;
    int chk;
    int ret;
    char protocol[L_URL], portstr[L_URL];
    char addrbuf[32];

    ret = 0;

#ifdef TRACE_HCACHE
    Trace("hcache_recv_response_addr_direct: name '%s'\n", name);
#endif

    if(name==NULL || *name=='\0') {
        return -1;
    }
    if(*qfd<0) {
        return -1;
    }

try_read:
    nr = read(*qfd, addrbuf, 32);
#ifdef TRACE_HCACHE
    Trace("  read %d bytes (%d)\n", nr, errno);
#endif
    if(nr<0) {
        if(errno==EINTR) {
            goto try_read;
        }
        Error("resolver negative reply, ret %d (%d)", nr, errno);
        ret = -4;
        goto try_close;
    }
    else
    if(nr>0) {
        addr->sin_family = AF_INET;
        memcpy(&(addr->sin_addr.s_addr), addrbuf, 4);
    }
    else {
        Error("empty resolver reply");

        ret = -5;
    }

try_close:
    shutdown(*qfd, 2);
    close(*qfd);
    *qfd = -1;

out:
    return ret;
}


int
hcache_seek_addr_incache(char *name, struct sockaddr_in* addr)
{
    int chk;
    int ret;
    char protocol[L_URL], portstr[L_URL];
    int qfd;
    char addrbuf[32];
    int i, h, s, pos, free_pos, min, min_pos;
    time_t now;

#ifdef TRACE_HCACHE
    Trace("hcache_got_addr_wcache: name '%s'\n", name);
#endif

    if(name==NULL || *name=='\0') {
        return -1;
    }

    ret = -1;


    /***
     *** check numeric name
     ***/
    if(*name>='0' && *name<='9') {
        register char *p;

        /* check, name is filled digits and/or dots */
        p = name;
        while(*p && ((*p>='0' && *p<='9') || *p=='.')) {
            p++;
        }

        if(!*p) {
            memset(addr, 0, sizeof(addr));
            addr->sin_family = AF_INET;
            addr->sin_port   = 0;
            addr->sin_addr.s_addr = inet_addr(name);

            ret = 0;
            goto out;
        }
    }


    /***
     *** check cache
     ***/

    h = (int)hhash((unsigned char*)name);
    now = time(NULL);

#ifdef TRACE_DETAIL_HCACHE
    fprintf(stderr, "    *** hash value %d\n", h);

    MUTEX_LOCK(&hcache_lock);
    for(i=0;i<HCACHE_NSLOT;i++) {
        fprintf(stderr, "%4d: %12d '%s'\n", i, slot_use[i], hname_slot[i]);
    }
    MUTEX_UNLOCK(&hcache_lock);
#endif

    MUTEX_LOCK(&hcache_lock);


    /*
     * Search same name, free position and minimum value postion in
     * once sweep.  However, if I check all entry, it will be liner search.
     * Then, this routine check in short range (quoater) per one sweep.
     */

    pos = -1;
    free_pos = -1;
    min = INT_MAX;
    min_pos = -1;
    i = 0;
    s = h;
    while(i<HCACHE_NSLOT/4) {
        if(hname_slot[s][0] == *name && strcmp(hname_slot[s], name)==0) {
            pos = s;
            break;
        }
        if(hname_slot[s][0] == '\0') {
            free_pos = s;
            break;
        }

        if(slot_use[s]<min) {
            min = slot_use[s];
            min_pos = s;
        }

        i++;
        s = (s+1)%HCACHE_NSLOT;
    }


    /*
     * found
     */
    if(pos>=0) {
#ifdef TRACE_HCACHE
        Trace("  found in hostname cache.\n");
#endif

        memcpy(addr, &haddr_slot[pos], sizeof(struct sockaddr_in));

/*
        slot_use[pos]++;
*/
        slot_use[pos] = (int)now;

        ret = 0;
        goto out;
    }

out:
    MUTEX_UNLOCK(&hcache_lock);
    return ret;
}



/* with cache */
int
hcache_set_addr_incache(char *name, struct sockaddr_in* addr)
{
    int chk;
    int ret;
    char protocol[L_URL], portstr[L_URL];
    int qfd;
    char addrbuf[32];
    int i, h, s, pos, free_pos, min, min_pos;
    time_t now;

#ifdef TRACE_HCACHE
    Trace("hcache_got_addr_incache: name '%s'\n", name);
#endif

    if(name==NULL || *name=='\0') {
        return -1;
    }

    ret = -1;


    /***
     *** check cache
     ***/

    h = (int)hhash((unsigned char*)name);
    now = time(NULL);

#ifdef TRACE_DETAIL_HCACHE
    fprintf(stderr, "    *** hash value %d\n", h);

    MUTEX_LOCK(&hcache_lock);
    for(i=0;i<HCACHE_NSLOT;i++) {
        fprintf(stderr, "%4d: %12d '%s'\n", i, slot_use[i], hname_slot[i]);
    }
    MUTEX_UNLOCK(&hcache_lock);
#endif

    MUTEX_LOCK(&hcache_lock);


    /*
     * Search same name, free position and minimum value postion in
     * once sweep.  However, if I check all entry, it will be liner search.
     * Then, this routine check in short range (quoater) per one sweep.
     */

    pos = -1;
    free_pos = -1;
    min = INT_MAX;
    min_pos = -1;
    i = 0;
    s = h;
    while(i<HCACHE_NSLOT/4) {
        if(hname_slot[s][0] == *name && strcmp(hname_slot[s], name)==0) {
            pos = s;
            break;
        }
        if(hname_slot[s][0] == '\0') {
            free_pos = s;
            break;
        }

        if(slot_use[s]<min) {
            min = slot_use[s];
            min_pos = s;
        }

        i++;
        s = (s+1)%HCACHE_NSLOT;
    }


    /*
     * found
     */
    if(pos>=0) {
#ifdef TRACE_HCACHE
        Trace("  found in hostname cache.\n");
#endif

#if 0
        memcpy(addr, &haddr_slot[pos], sizeof(struct sockaddr_in));
#endif
        memcpy(&haddr_slot[pos], addr, sizeof(struct sockaddr_in));

/*
        slot_use[pos]++;
*/
        slot_use[pos] = (int)now;

        ret = 0;
        goto out;
    }


#ifdef TRACE_HCACHE
    /*
    Trace("  not found in hostname cache, retrieve from other system.\n");
    */
    Trace("  still it is not exist\n");
#endif

    /*
     * not found
     */
    if(free_pos==-1) {
        if(min_pos==-1) {
            free_pos = h;
        }
        else {
            free_pos = min_pos;
        }
    }


    /* cachable */
    if(strlen(name)<L_HOST-1) {
        strncpy(hname_slot[free_pos], name, L_HOST);
        memcpy(&haddr_slot[free_pos], addr, sizeof(struct sockaddr_in));
        slot_use[free_pos] = (int)now;
    }

    ret = 0;


out:
    MUTEX_UNLOCK(&hcache_lock);
    return ret;
}


/* with cache */
int
hcache_got_addr_wcache(char *name, struct sockaddr_in* addr)
{
    int chk;
    int ret;
    char protocol[L_URL], portstr[L_URL];
    int qfd;
    char addrbuf[32];
    int i, h, s, pos, free_pos, min, min_pos;
    time_t now;

#ifdef TRACE_HCACHE
    Trace("hcache_got_addr_wcache: name '%s'\n", name);
#endif

    if(name==NULL || *name=='\0') {
        return -1;
    }

    ret = -1;


    /***
     *** check cache
     ***/

    h = (int)hhash((unsigned char*)name);
    now = time(NULL);

#ifdef TRACE_DETAIL_HCACHE
    fprintf(stderr, "    *** hash value %d\n", h);

    MUTEX_LOCK(&hcache_lock);
    for(i=0;i<HCACHE_NSLOT;i++) {
        fprintf(stderr, "%4d: %12d '%s'\n", i, slot_use[i], hname_slot[i]);
    }
    MUTEX_UNLOCK(&hcache_lock);
#endif

    MUTEX_LOCK(&hcache_lock);


    /*
     * Search same name, free position and minimum value postion in
     * once sweep.  However, if I check all entry, it will be liner search.
     * Then, this routine check in short range (quoater) per one sweep.
     */

    pos = -1;
    free_pos = -1;
    min = INT_MAX;
    min_pos = -1;
    i = 0;
    s = h;
    while(i<HCACHE_NSLOT/4) {
        if(hname_slot[s][0] == *name && strcmp(hname_slot[s], name)==0) {
            pos = s;
            break;
        }
        if(hname_slot[s][0] == '\0') {
            free_pos = s;
            break;
        }

        if(slot_use[s]<min) {
            min = slot_use[s];
            min_pos = s;
        }

        i++;
        s = (s+1)%HCACHE_NSLOT;
    }


    /*
     * found
     */
    if(pos>=0) {
#ifdef TRACE_HCACHE
        Trace("  found in hostname cache.\n");
#endif

        memcpy(addr, &haddr_slot[pos], sizeof(struct sockaddr_in));

/*
        slot_use[pos]++;
*/
        slot_use[pos] = (int)now;

        ret = 0;
        goto out;
    }


#ifdef TRACE_HCACHE
    Trace("  not found in hostname cache, retrieve from other system.\n");
#endif

    /*
     * not found
     */
    if(free_pos==-1) {
        if(min_pos==-1) {
            free_pos = h;
        }
        else {
            free_pos = min_pos;
        }
    }

    chk = hcache_got_addr_direct(name, addr);

    if(chk==0) {
        /* success */

        /* cachable */
        if(strlen(name)<L_HOST-1) {
            strncpy(hname_slot[free_pos], name, L_HOST);
            memcpy(&haddr_slot[free_pos], addr, sizeof(struct sockaddr_in));
/*
            slot_use[free_pos] = 1;
*/
            slot_use[free_pos] = (int)now;
        }

        ret = 0;
    }
    else {
        Error("cannot get hostname [%d]", chk);
        ret = chk;
    }


out:
    MUTEX_UNLOCK(&hcache_lock);
    return ret;
}



int
hcache_got_addr(char *name, struct sockaddr_in* addr)
{
    int ret;
    struct timeval before, after;
    long cv;


    if(hres_stat.count>STAT_LIFETIME) {
        sim_stat_reset(&hres_stat);
    }


    gettimeofday(&before, NULL);
    ret = hcache_got_addr_wcache(name, addr);
    gettimeofday(&after, NULL);


#if 0
    cv = ((after.tv_sec%3600)*1000000+after.tv_usec)
            -((before.tv_sec%3600)*1000000+before.tv_usec) ;
#endif
    cv = KOTETU_SUB_TIME(after, before);

    sim_stat_add_value(&hres_stat, cv);

#if 0
    if(hres_stat.count%STAT_REPORT==0) {
        sim_stat_show_lock(&hres_stat, stderr);
    }
#endif


    return ret;
}
